package org.LiveGraph.dataCache;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.LiveGraph.settings.DataFileSettings;
import org.LiveGraph.settings.ErrorWhileSettingHasChangedProcessingException;
import org.LiveGraph.settings.ObservableSettings;
import org.LiveGraph.settings.SettingsObserver;


/**
 * An object of this class is used to triger updates from a data input stream
 * into a {@link DataCache} at regular intervals.
 * 
 * <p style="font-size:smaller;">This product includes software developed by the
 *    <strong>LiveGraph</strong> project and its contributors.<br />
 *    (<a href="http://www.live-graph.org" target="_blank">http://www.live-graph.org</a>)<br />
 *    Copyright (c) 2007 G. Paperin.<br />
 *    All rights reserved.
 * </p>
 * <p style="font-size:smaller;">File: UpdateInvoker.java</p> 
 * <p style="font-size:smaller;">Redistribution and use in source and binary forms, with or
 *    without modification, are permitted provided that the following terms and conditions are met:
 * </p>
 * <p style="font-size:smaller;">1. Redistributions of source code must retain the above
 *    acknowledgement of the LiveGraph project and its web-site, the above copyright notice,
 *    this list of conditions and the following disclaimer.<br />
 *    2. Redistributions in binary form must reproduce the above acknowledgement of the
 *    LiveGraph project and its web-site, the above copyright notice, this list of conditions
 *    and the following disclaimer in the documentation and/or other materials provided with
 *    the distribution.<br />
 *    3. All advertising materials mentioning features or use of this software or any derived
 *    software must display the following acknowledgement:<br />
 *    <em>This product includes software developed by the LiveGraph project and its
 *    contributors.<br />(http://www.live-graph.org)</em><br />
 *    4. All advertising materials distributed in form of HTML pages or any other technology
 *    permitting active hyper-links that mention features or use of this software or any
 *    derived software must display the acknowledgment specified in condition 3 of this
 *    agreement, and in addition, include a visible and working hyper-link to the LiveGraph
 *    homepage (http://www.live-graph.org).
 * </p>
 * <p style="font-size:smaller;">THIS SOFTWARE IS PROVIDED &quot;AS IS&quot;, WITHOUT WARRANTY
 *    OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND  NONINFRINGEMENT. IN NO EVENT SHALL
 *    THE AUTHORS, CONTRIBUTORS OR COPYRIGHT  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING  FROM, OUT OF OR
 *    IN CONNECTION WITH THE SOFTWARE OR THE USE OR  OTHER DEALINGS IN THE SOFTWARE.
 * </p>
 * 
 * @author Greg Paperin (<a href="http://www.paperin.org" target="_blank">http://www.paperin.org</a>)
 * @version {@value org.LiveGraph.LiveGraph#version}
 */
public class UpdateInvoker implements Runnable, SettingsObserver {

/**
 * How long to sleep for when updates are to be invoked automatically.
 */
private static final long defaultTimeTickLength = 500; // milliseconds 

/**
 * How long to sleep for when updates are to be invoked manually.
 */
private static final long restingTimeTickLength = 2000; // milliseconds


/**
 * The data reader for the input stream.
 */
private DataStreamToCacheReader dataReader = null;

/**
 * Cache to hold the data.
 */
private DataCache dataCache = null;

/**
 * Data file from which to update.
 */
private File dataFile = null;

/**
 * Whether cache must be reset before each read.
 */
private boolean dontCache = false;


/**
 * Interval between updates in milliseconds.
 */
private long interval = -1;

/**
 * Whether an update is currently running. 
 */
private boolean updateInProgress = false;

/**
 * Whether the invoker thread should wind up at the next possibility.
 */
private boolean mustQuit = false;

/**
 * The sleep length betweek ticks at which this invoker can react to events and send
 * mesages.
 */
private long timeTickLength = restingTimeTickLength;

/**
 * Remaining milliseconds till the next update.
 */
private long remainingMillis = -1;

/**
 * System milliseconds at last update.
 */
private long lastUpdateTime = 0;

/**
 * Milliseconds since last update.
 */
private long sinceUpdateTime = 0;

/**
 * Thread actually performing the update.
 */
private Thread fileReadingThread = null;


/**
 * List of observers.
 */
private List<UpdateInvokerObserver> observers = null;

/**
 * Constructs a new invoker.
 */
public UpdateInvoker() {
	dataReader = null;
	dataCache = null;
	dataFile = null;
	dontCache = false;
	
	interval = -1;
	updateInProgress = false;
	mustQuit = false;
	timeTickLength = restingTimeTickLength;
	remainingMillis = -1;
	lastUpdateTime = 0;
	sinceUpdateTime = 0;
	fileReadingThread = null;
	
	observers = new ArrayList<UpdateInvokerObserver>();
}

/**
 * Sets the file from which the next update will be read and resets the data cache.
 * 
 * @param file File from which to read the data from now on.
 * @throws FileNotFoundException If no such file exists.
 * @throws IllegalStateException If no valid data cache is set.
 */
public synchronized void setDataFile(File file) throws FileNotFoundException {
	if (null == file)
		throw new NullPointerException("Cannot read data from a null file");
	
	if (null == dataCache)
		throw new IllegalStateException("Must set dataCache before setting an input stream");
	
	dataFile = file;
	createDataReader();	
}

/**
 * Sets the file from which the next update will be read and resets the data cache.
 * 
 * @param fileName File from which to read the data from now on.
 * @throws FileNotFoundException If no such file exists.
 * @throws IllegalStateException If no valid data cache is set.
 */
public synchronized void setDataFile(String fileName) throws FileNotFoundException {
	
	if (null == fileName)
		throw new NullPointerException("Cannot read data from a null file name");
	
	setDataFile(new File(fileName));
}

/**
 * Sets whether the cache should be reset before each update.
 * @param state Indicates whether the cache should be reset before each update.
 */
public synchronized void setDontCacheData(boolean state) {
	dontCache = state;
}

/**
 * Specifies the data cache to which the data read on the next update will be written.
 * 
 * @param cache The data cache to use from now on. 
 */
public synchronized void setDataCache(DataCache cache) {
	if (null == cache)
		throw new NullPointerException("Cannot read data into a null cache");
	dataCache = cache;
}

/**
 * Sets the length of the interval between automatic data updates in milliseconds.
 * If {@code interval <= 0} the update will not be triggered automatically.
 * 
 * @param interval The length of the interval between automatic data updates in milliseconds
 * (if {@code interval <= 0} the update will not be triggered automatically).
 */
public synchronized void setInterval(long interval) {
	if (this.interval <= 0 && interval > 0)
		this.lastUpdateTime = 0;
	this.interval = interval;
	this.timeTickLength = interval <= 0 ? restingTimeTickLength : defaultTimeTickLength;
}

/**
 * Creates a reader on the currently set data input file and resets the cache.
 * 
 * @throws FileNotFoundException If the input file does not exist.
 */
private synchronized void createDataReader() throws FileNotFoundException {
	
	if (null == dataFile)
		return;
	
	if (0 == dataFile.getPath().length())
		return;
	
	if (!dataFile.exists())
		throw new FileNotFoundException("Data source file does not exist (" + dataFile.getPath() + ")");
	
	if (null != dataReader)
		closeReader();
	
	dataCache.resetCache();
	FileInputStream ins = new FileInputStream(dataFile);	
	dataReader = new DataStreamToCacheReader(ins, dataCache);
}

/**
 * The length of the interval between data updates.
 * If {@code interval <= 0} the update will not be triggered automatically.
 * @return The length of the interval between automatic data updates in milliseconds;
 * a value {@code interval <= 0} indicated that no updates will be triggered automatically.
 */
public long getInterval() {
	return interval;
}

/**
 * Executes the next update from the current input file into the current data cache.
 * 
 * @return {@code true} if the update was started successfully, {@code false} if the
 * update could not be started (possibly, because a previous update was still in progress).
 */
public synchronized boolean update() {
	
	if (isUpdateInProgress())
		return false;
	
	setUpdateInProgress(true);
	notifyObserversUpdateStarted();
	
	fileReadingThread = new Thread(new Runnable() {
		public void run() {
			
			String errorMsg = null;
			try {
				
				if (dontCache || null == dataReader)
					createDataReader();
				
				if (null == dataReader) {
					notifyObserversUpdateFinished("Data source not set or data source not valid");
					lastUpdateTime = System.currentTimeMillis();
					setUpdateInProgress(false);
					return;
				}
					
				dataCache.startDelayEvents();
				dataReader.readFromStream();
				dataCache.fireDelayedEvents();
				
				if (dontCache)
					closeReader();
				
			} catch(Exception e) {				
				errorMsg = e.getMessage() + "(" + e.getClass().getSimpleName() + ")";
			}
			
			Thread.yield();
			notifyObserversUpdateFinished(errorMsg);
			lastUpdateTime = System.currentTimeMillis();
			setUpdateInProgress(false);
		}
	}, "File reading thread");
	
	fileReadingThread.start();
	
	return true;
}

/**
 * Add an observer to this invoker.
 * 
 * @param observer Observer to add.
 * @return Whether the observer was really added because it was not there yet.
 */
public boolean addObserver(UpdateInvokerObserver observer) {
	if (hasObserver(observer))
		return false;
	return observers.add(observer);
}

/**
 * Check for the specified observer.
 * 
 * @param observer An observer.
 * @return Whether the specified observer is in the observer list.
 */
public boolean hasObserver(UpdateInvokerObserver observer) {
	return observers.contains(observer);	
}

/**
 * Removes an observer.
 * 
 * @param observer An observer.
 * @return Whether the observer was successfully removed.
 */
public boolean removeObserver(UpdateInvokerObserver observer) {
	return observers.remove(observer);
}

/**
 * Count observers.
 * 
 * @return Current number of observers.
 */
public int countObervers() {
	return observers.size();
}

/**
 * Notifies this invoker's observers that this observer has waken up to process events.
 * This gives listeners displaying various information about this invoker a chance to
 * update their state.
 */
public void notifyObserversTimerTick() {
	for (UpdateInvokerObserver observer : observers)
		observer.timerTick(this);
}

/**
 * Notifies this invoker's observers that an update was triggered.
 */
public void notifyObserversUpdateStarted() {
	for (UpdateInvokerObserver observer : observers)
		observer.updateStarted(this);
}

/**
 * Notifies this invoker's observers that an update has finished.
 *  
 * @param errorMsg Describes any problems during the update. If there were no problems
 * {@code errorMsg} will be {@code null}.
 */
public void notifyObserversUpdateFinished(String errorMsg) {
	for (UpdateInvokerObserver observer : observers)
		observer.updateFinished(this, errorMsg);
}

/**
 * Used to notify this invoker that is must stop running at the next possibility.
 * 
 * @param val Whether this invoker should stop running at the next possibility.
 */
public synchronized void setMustQuit(boolean val) {
	this.mustQuit = val;
}

/**
 * Time to next update.
 * 
 * @return Number of milliseconds left until the next update.
 */
public long getRemainingMillis() {
	return remainingMillis;
}

/**
 * Whether an update is currently running.
 * @return Whether an update is currently running.
 */
public boolean isUpdateInProgress() {
	return updateInProgress;
}

/**
 * Sets the internal {@code updateInProgress} state.
 * @param state The new state.
 */
private synchronized void setUpdateInProgress(boolean state) {
	updateInProgress = state;
}

/**
 * Closes the current data reader.
 */
private void closeReader() {
	try {
		dataReader.close();
	} catch (IOException e) {
		e.printStackTrace();
	}
	dataReader = null;
}

/**
 * Winds up the operations by closing the current data reader.
 */
private void tidyUp() {
	if (null != dataReader) {
		closeReader();
	}
}

/**
 * Send the this invoker to sleep for {@code timeTickLength} milliseconds.
 * When it wakes it, internal time state is updated an the observers notified.
 */
private void timeTick() {
	Thread.yield();
	try { Thread.sleep(timeTickLength); } catch (InterruptedException e) {}
	
	sinceUpdateTime = System.currentTimeMillis() - lastUpdateTime;
	remainingMillis = interval <= 0 ? -1 : Math.max(0, interval - sinceUpdateTime);
	
	notifyObserversTimerTick();
	Thread.yield();
}

/**
 * Main invoker loop:
 * call {@link #timeTick()};
 * if it is time for the next update, call {@link #update()};
 * call {@link #timeTick()} again and continue the loop until {@link #mustQuit} is set to true;
 * call {@link #tidyUp()} before quitting. 
 */
public void run() {
		
	while (!mustQuit) {
		
		timeTick();
		
		if (sinceUpdateTime >= interval && interval > 0)
			update();
	}
	
	tidyUp();
}

/**
 * Dispatches settings change events.
 */
public void settingHasChanged(ObservableSettings settings, Object info) {
	
	if (null == info || null == settings)
		return;
	
	if ((settings instanceof DataFileSettings) && (info instanceof String)) {
		settingHasChanged((DataFileSettings) settings, (String) info);
		return;
	}
}

/**
 * When the application's data file settings change which method is called in order
 * to update the internal state accordingly.
 * 
 * @param settings Application's data file settings.
 * @param info Describes the change event.
 */
public void settingHasChanged(DataFileSettings settings, String info) {
	
	if (null == info || null == settings)
		return;
	
	if (info.equals("DataFile")) {
		DataFileSettings_DataFile(settings);
		return;
	}
	
	if (info.equals("UpdateFrequency")) {
		DataFileSettings_UpdateFrequency(settings);
		return;
	}
	
	if (info.equals("DoNotCacheData")) {
		DataFileSettings_DoNotCacheData(settings);
		return;
	}
	
	if (info.equals("ShowOnlyTailData")) {
		DataFileSettings_ShowOnlyTailData(settings);		
		return;
	}
	
	if (info.equals("load")) {
		ErrorWhileSettingHasChangedProcessingException ex = null;
		
		try {
			DataFileSettings_DataFile(settings);
		} catch (ErrorWhileSettingHasChangedProcessingException e) {
			if (null == ex) ex = e;
		}
		
		try {
			DataFileSettings_UpdateFrequency(settings);
		} catch (ErrorWhileSettingHasChangedProcessingException e) {
			if (null == ex) ex = e;
		}
		
		try {
			DataFileSettings_DoNotCacheData(settings);
		} catch (ErrorWhileSettingHasChangedProcessingException e) {
			if (null == ex) ex = e;
		}
		
		try {
			DataFileSettings_ShowOnlyTailData(settings);
		} catch (ErrorWhileSettingHasChangedProcessingException e) {
			if (null == ex) ex = e;
		}
		
		if (null != ex) {
			ex.setDontBreakObserverNotification(true);
			throw ex;
		}
		
		return;
	}
}

/**
 * Adjusts the file used by this invoker in response to a corresponding change in the
 * data file settings.
 * 
 * @param settings The data file settigs.
 */
private void DataFileSettings_DataFile(DataFileSettings settings) {
	try {
		setDataFile(settings.getDataFile());
		update();
	} catch (FileNotFoundException e1) {
		throw new ErrorWhileSettingHasChangedProcessingException(e1);
	} catch (NullPointerException e2) {
		throw new ErrorWhileSettingHasChangedProcessingException(e2);			
	}
}

/**
 * Adjusts the update frequency used by this invoker in response to a corresponding change in the
 * data file settings.
 * 
 * @param settings The data file settigs.
 */
private void DataFileSettings_UpdateFrequency(DataFileSettings settings) {
	setInterval(settings.getUpdateFrequency());
}

/**
 * Adjusts the cache resetting option used by this invoker in response to a corresponding change in the
 * data file settings.
 * 
 * @param settings The data file settigs.
 */
private void DataFileSettings_DoNotCacheData(DataFileSettings settings) {
	setDontCacheData(settings.getDoNotCacheData());
}

/**
 * Adjusts the cache mode used by this invoker in response to a corresponding change in the
 * data file settings.
 * 
 * @param settings The data file settigs.
 */
private void DataFileSettings_ShowOnlyTailData(DataFileSettings settings) {
	if (null == dataCache)
		return;
	
	boolean onlyTale = settings.getShowOnlyTailData();
	
	if (dataCache.getCacheMode() == DataCache.CacheMode.CacheAllData && !onlyTale)
		return;
	
	if (dataCache.getCacheMode() == DataCache.CacheMode.CacheTailData && onlyTale)
		return;
	
	dataCache.resetData(onlyTale ? DataCache.CacheMode.CacheTailData : DataCache.CacheMode.CacheAllData);
	try {
		createDataReader();
		update();
	} catch (FileNotFoundException e) {
		throw new ErrorWhileSettingHasChangedProcessingException(e);
	}
}

}
