package org.LiveGraph.plot;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.LiveGraph.LiveGraph;
import org.LiveGraph.dataCache.CacheObserver;
import org.LiveGraph.dataCache.DataCache;
import org.LiveGraph.dataCache.DataSeries;
import org.LiveGraph.dataCache.DataSet;
import org.LiveGraph.settings.DataSeriesSettings;
import org.LiveGraph.settings.GraphSettings;
import org.LiveGraph.settings.ObservableSettings;
import org.LiveGraph.settings.SettingsObserver;
import org.LiveGraph.settings.DataSeriesSettings.TransformMode;
import org.LiveGraph.settings.GraphSettings.HGridType;
import org.LiveGraph.settings.GraphSettings.VGridType;
import org.LiveGraph.settings.GraphSettings.XAxisType;


/**
 * This class handles the conversion of the cached data to a screen image and the
 * drawing of the image on a {@code Graphics} object.
 * <br />
 * This class uses an {@code AffineTransform} object to convert the data held in the
 * cache to a data plot in screen coordinates. In order to keep the {@code AffineTransform}
 * object appropriate for the current display at all times a plotter listens to
 * various {@link DataCache} and {@link ObservableSettings} events; in addition it offers
 * a {@link #setScreenSize(int, int)}-method which must be called each time when the
 * canvas-panel that uses the plotter changes its size.
 * <br />
 * Whenever the {@link #dataCache} changes, a plotter uses the current {@link #datScrTransform}
 * object to convert the data from the cache into a plot in screen coordinates according to
 * the current global graph- and series-settings. The screen data obtained this way is locally
 * cached in the {@link #screenDataBuffer} array. This way the data does not need to be
 * re-computed each time the plot must be drawn on the screen.
 * <br />
 * In this version the plotter handles data values transformations required by the display
 * options (if any) on the fly. If new options should be added to theinterface, this mechanism
 * should be replaces by a more flexible solution.
 * 
 * <p style="font-size:smaller;">This product includes software developed by the
 *    <strong>LiveGraph</strong> project and its contributors.<br />
 *    (<a href="http://www.live-graph.org" target="_blank">http://www.live-graph.org</a>)<br />
 *    Copyright (c) 2007 G. Paperin.<br />
 *    All rights reserved.
 * </p>
 * <p style="font-size:smaller;">File: Plotter.java</p> 
 * <p style="font-size:smaller;">Redistribution and use in source and binary forms, with or
 *    without modification, are permitted provided that the following terms and conditions are met:
 * </p>
 * <p style="font-size:smaller;">1. Redistributions of source code must retain the above
 *    acknowledgement of the LiveGraph project and its web-site, the above copyright notice,
 *    this list of conditions and the following disclaimer.<br />
 *    2. Redistributions in binary form must reproduce the above acknowledgement of the
 *    LiveGraph project and its web-site, the above copyright notice, this list of conditions
 *    and the following disclaimer in the documentation and/or other materials provided with
 *    the distribution.<br />
 *    3. All advertising materials mentioning features or use of this software or any derived
 *    software must display the following acknowledgement:<br />
 *    <em>This product includes software developed by the LiveGraph project and its
 *    contributors.<br />(http://www.live-graph.org)</em><br />
 *    4. All advertising materials distributed in form of HTML pages or any other technology
 *    permitting active hyper-links that mention features or use of this software or any
 *    derived software must display the acknowledgment specified in condition 3 of this
 *    agreement, and in addition, include a visible and working hyper-link to the LiveGraph
 *    homepage (http://www.live-graph.org).
 * </p>
 * <p style="font-size:smaller;">THIS SOFTWARE IS PROVIDED &quot;AS IS&quot;, WITHOUT WARRANTY
 *    OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND  NONINFRINGEMENT. IN NO EVENT SHALL
 *    THE AUTHORS, CONTRIBUTORS OR COPYRIGHT  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING  FROM, OUT OF OR
 *    IN CONNECTION WITH THE SOFTWARE OR THE USE OR  OTHER DEALINGS IN THE SOFTWARE.
 * </p>
 * 
 * @author Greg Paperin (<a href="http://www.paperin.org" target="_blank">http://www.paperin.org</a>)
 * @version {@value org.LiveGraph.LiveGraph#version}
 */
public class Plotter implements CacheObserver, SettingsObserver {

/**
 * Vertical margin.
 */
private static final int VMARGIN = 20;

/**
 * Horisiontal margin.
 */
private static final int HMARGIN = 20;

/*
 * Minimum plot size.
 */
private static final Dimension minScreenSize = new Dimension(150, 100);

/**
 * Y axis colour.
 */
private static final Color VAXIS_COL = Color.BLACK;

/**
 * X axis colour.
 */
private static final Color HAXIS_COL = Color.BLACK;

/**
 * Label font size.
 */
private static final int FONT_SIZE = 9;

/**
 * Gap between axes labels.
 */
private static final int AXES_LBL_GAP = 100;

/**
 * Size of the scale marks on the axes.
 */
private static final int AXES_MARKS_SIZE = 4;

/**
 * Radius for datapoints marks on small graphs.
 */
private static final int DATAPOINT_RAD = 3;

/**
 * The minimum distance between grid lines (in pixels).
 */
private static final int MIN_GRIDLINE_DIST = 3;


/**
 * The data cache. 
 */
private DataCache dataCache = null;

/**
 * Data series settings.
 */
private DataSeriesSettings seriesSetts = null;

/**
 * Graph settings.
 */
private GraphSettings graphSetts = null;


/**
 * Whether anythig at all is to be displayed.
 */
private boolean showAtLeastOneSeries = false;

/**
 * Buffers the screen coordinates of the graphs.
 */
private SeriesScreenData[] screenDataBuffer = null;

/**
 * Buffers the x coordinates.
 */
private double[] xCoordinates = null;

/**
 * Used for sorting points by x values. 
 */
private PointsByIndexComparator pointsByIndexComparator = null; 


/**
 * Viewable area in data coordinates.
 */
private Rectangle2D.Double dataViewport = null;

/**
 * Screen size in pixels.
 */
private Dimension screenSize = null;

/**
 * Data space to screen space transformation.
 */
private AffineTransform datScrTransform = null;


/**
 * Whether screen data computation is in progress.
 */
private boolean dataComputationRunning = false;

/**
 * Whether screen data computation is in progress.
 */
private boolean pointHighlightComputationRunning = false;

/**
 * Whether the next change of h-grid settings was initiated by this plotter and should
 * therefore be ignored by the plotter's handler.
 */
private boolean selfSettingHGridSize = false;

/**
 * The h-grid size get by a settings change that was not initiated by this
 * plotter itself.
 */
private double userHGridStep = Double.NaN;

/**
 * The actual h-grid step after the consideration of plot size.
 */
private double hGridStep = Double.NaN;

/**
 * Whether the next change of v-grid settings was initiated by this plotter and should
 * therefore be ignored by the plotter's handler.
 */
private boolean selfSettingVGridSize = false;

/**
 * The v-grid size get by a settings change that was not initiated by this
 * plotter itself.
 */
private double userVGridStep = Double.NaN;

/**
 * The actual v-grid step after the consideration of plot size.
 */
private double vGridStep = Double.NaN;


/**
 * Whether dara points close to the mouse position should be highlighted. 
 */
private boolean highlightPoints = true;


/**
 * Creates a plotter for the data held in the specified cache.
 * @param dataCache Cache holding the data to plot.
 */
public Plotter(DataCache dataCache) {
	
	if (null == dataCache)
		throw new NullPointerException("Plotter cannot act on a null cache"); 
	
	this.dataCache = dataCache;
	this.initGlobalParameters();
	
	this.resetScreenDataBuffer();
	this.xCoordinates = new double[DataCache.CACHE_SIZE];
	this.pointsByIndexComparator = new PointsByIndexComparator();
	
	this.dataViewport = new Rectangle2D.Double();
	this.screenSize = new Dimension();
	this.datScrTransform = new AffineTransform();
	
	this.dataComputationRunning = false;
	this.pointHighlightComputationRunning = false;
	
	this.selfSettingHGridSize = false;
	this.userHGridStep = graphSetts.getHGridSize();
	this.hGridStep = graphSetts.getHGridSize();
	
	this.selfSettingVGridSize = false;
	this.userVGridStep = graphSetts.getVGridSize();
	this.vGridStep = graphSetts.getVGridSize();
	
	this.computeGridSteps();
	
	this.highlightPoints = true;
}

/**
 * Used by the constructor to initialise global settings references.
 */
private void initGlobalParameters() {
	this.seriesSetts = LiveGraph.application().getDataSeriesSettings();
	this.graphSetts = LiveGraph.application().getGraphSettings();
}


/**
 * Gets whether the screen area is large enough to paint the graph.
 * 
 * @return {@code true} iff the screen area is large enough to paint the graph.
 */
public boolean screenTooSmall() {
	return (minScreenSize.height > screenSize.height || minScreenSize.width > screenSize.width);
}

/**
 * Gets whether at least one series is to be plotted.
 * 
 * @return {@code true} if at seast one data series should be plotted, {@code false} otherwise.
 */
public boolean getShowAtLeastOneSeries() {
	return this.showAtLeastOneSeries;
}

/**
 * Paints the previously computed graphs along with the axes, labels, grids and so on to the
 * specified graphics context.
 * @param g Paint context.
 */
public void paint(Graphics g) {
		
	// If screen is to small, just paint a message:	
	if (screenTooSmall()) {
		g.setColor(Color.BLACK);	
		g.setFont(new Font(g.getFont().getName(), g.getFont().getStyle(), FONT_SIZE));
		FontMetrics fMetrics = g.getFontMetrics();
		g.drawString("LiveGraph " + LiveGraph.version, 2, fMetrics.getHeight() + 2);
		g.drawString("Enlarge this window to see the plot.", 2, 2 * fMetrics.getHeight() + 4);
		return;
	}

	
	// If there is nothing to show, just print a message:	
	if (!showAtLeastOneSeries) {
		g.setColor(Color.BLACK);
		g.setFont(new Font(g.getFont().getName(), g.getFont().getStyle(), FONT_SIZE));
		FontMetrics fMetrics = g.getFontMetrics();
		g.drawString("LiveGraph " + LiveGraph.version, 2, fMetrics.getHeight() + 2);
		g.drawString("No data to display.", 2, 2 * fMetrics.getHeight() + 4);
		return;
	}
	
	
	// If data computation is running, do not do anything:	
	if (dataComputationRunning)
		return;
	
	// Now do the actual painting:
	paintGrids(g);
	paintAxes(g);
	paintData(g);
	
} // public void paint(Graphics g)


/**
 * Paints the grid.
 * @param g The graphics context.
 */
private void paintGrids(Graphics g) {
	
	// Plot horizontal grid:
	
	if (HGridType.HGrid_Simple == graphSetts.getHGridType()) {
		
		g.setColor(graphSetts.getHGridColour());
		
		double dataViewportBottom = dataViewport.y - dataViewport.height;
		double gy = dataViewportBottom + hGridStep - (dataViewportBottom % hGridStep);
		
		Point2D.Double p1 = new Point2D.Double();
		Point2D.Double p2 = new Point2D.Double();
		
		while (gy <= dataViewport.y) {
			
			p1.setLocation(dataViewport.x, gy);
			p2.setLocation(dataViewport.x + dataViewport.width, gy);
			datScrTransform.transform(p1, p1);
			datScrTransform.transform(p2, p2);
			g.drawLine((int) p1.x, (int) p1.y, (int) p2.x, (int) p2.y);
			gy += hGridStep;
		}
	}
	
	
	// Plot vertical grid if it is alligned at x-axis units:
	
	if (VGridType.VGrid_XAUnitAligned == graphSetts.getVGridType()) {
		
		g.setColor(graphSetts.getVGridColour());
		
		double gx = dataViewport.x + vGridStep - (dataViewport.x % vGridStep);
		
		Point2D.Double p1 = new Point2D.Double();
		Point2D.Double p2 = new Point2D.Double();
		
		while (gx <= dataViewport.x + dataViewport.width) {
			
			p1.setLocation(gx, dataViewport.y);
			p2.setLocation(gx, dataViewport.y - dataViewport.height);
			datScrTransform.transform(p1, p1);
			datScrTransform.transform(p2, p2);
			g.drawLine((int) p1.x, (int) p1.y, (int) p2.x, (int) p2.y);
			gx += vGridStep;
		}
	}
	
	
	// Plot vertical grid if it is alligned at dataset file indices:
	
	// Get any (e.g. the first) data series which will be drawn:	
	SeriesScreenData firstSeriesVisible = null;
	for (int s = 0; s < screenDataBuffer.length; s++) {
		if (screenDataBuffer[s].doShow) {
			firstSeriesVisible = screenDataBuffer[s];
			break;
		}
	}
	
	if (VGridType.VGrid_DSNumAligned == graphSetts.getVGridType()) {
		
		g.setColor(graphSetts.getVGridColour());
		
		int gy1 = VMARGIN;
		int gy2 = VMARGIN + screenSize.height;
	
		int curDSInd, gx;
		int lastDSInd = firstSeriesVisible.dsIndices[0];
		for (int p = 0; p < firstSeriesVisible.plotPoints; p++) {
			
			curDSInd = firstSeriesVisible.dsIndices[p];
			if (curDSInd - lastDSInd >= vGridStep) {				
				gx = (int) firstSeriesVisible.points[p].x;
				g.drawLine(gx, gy1, gx, gy2);
				lastDSInd = curDSInd;				
			}
			
		}		
	}
} // private void paintGrids


/**
 * Paints the coordinate axes.
 * @param g The graphics context.
 */
private void paintAxes(Graphics g) {
	
	// Setup font:
	
	Font font = g.getFont();
	g.setFont(new Font(font.getName(), font.getStyle(), FONT_SIZE));
	FontMetrics fMetrics = g.getFontMetrics();
	
	// Plot horisontal axis:
	
	int xAxisY = VMARGIN + screenSize.height;
	Point sph = new Point();
	
	g.setColor(HAXIS_COL);
	g.drawLine(HMARGIN / 2, xAxisY, HMARGIN * 3 / 2 + screenSize.width, xAxisY);
	
	int sx = HMARGIN;
	while(true) {
		sph.setLocation(sx, xAxisY);
		Point2D.Double dp = screenToDataPoint(sph);
		String lbl = String.format("%.3f", dp.x);
		if (sx + 2 + fMetrics.stringWidth(lbl) + 2 > HMARGIN * 2 + screenSize.width)
			break;
		if (HMARGIN < sx)
			g.drawLine(sx, xAxisY - AXES_MARKS_SIZE / 2, sx, xAxisY + AXES_MARKS_SIZE / 2);				
		g.drawString(lbl, sx + 2, xAxisY + AXES_MARKS_SIZE / 2 + fMetrics.getHeight());
		sx += AXES_LBL_GAP;
	}
	
	
	// Plot vertical axis:
	
	int yAxisX = HMARGIN;
	Point spv = new Point();
	
	g.setColor(VAXIS_COL);
	g.drawLine(yAxisX, VMARGIN * 3 / 2 + screenSize.height, yAxisX, VMARGIN / 2);
	
	int sy = VMARGIN + screenSize.height;
	while(true) {
		spv.setLocation(yAxisX, sy);
		Point2D.Double dp = screenToDataPoint(spv);
		String lbl = String.format("%.3f", dp.y);
		if (sy - 2 - fMetrics.getHeight() - 2 < 0)
			break;
		if (VMARGIN + screenSize.height > sy)
			g.drawLine(yAxisX - AXES_MARKS_SIZE / 2, sy, yAxisX + AXES_MARKS_SIZE / 2, sy);				
		g.drawString(lbl, yAxisX + AXES_MARKS_SIZE / 2 + 2, sy - 2);
		sy -= AXES_LBL_GAP;
	}
	
} // private void paintAxes


/**
 * Paints the data series.
 * @param g The graphics context.
 */
private void paintData(Graphics g) {
	
	// Get any (e.g. the first) data series which will be drawn:	
	SeriesScreenData firstSeriesVisible = null;
	for (int s = 0; s < screenDataBuffer.length; s++) {
		if (screenDataBuffer[s].doShow) {
			firstSeriesVisible = screenDataBuffer[s];
			break;
		}
	}
	
	// Plot data:	
	boolean drawPoints = true;
	if (0 < firstSeriesVisible.plotPoints)
		drawPoints = (screenSize.width / firstSeriesVisible.plotPoints > 4 * DATAPOINT_RAD);
	
	SeriesScreenData series = null;
	for (int s = 0; s < screenDataBuffer.length; s++) {
		
		series = screenDataBuffer[s];
		if (!series.doShow)
			continue;
		
		g.setColor(series.colour);
		
		Point2D.Double[] points = series.points;
		int x1 = (int) points[0].x;
		int y1 = (int) points[0].y;
		int x2, y2;
		boolean connect = true;
		for (int p = 0; p < series.plotPoints; p++) {	
			if (Double.isNaN(points[p].x) || Double.isNaN(points[p].y)
					|| Double.isInfinite(points[p].x) || Double.isInfinite(points[p].y)) {
				connect = false;
				continue;				
			}			
			x2 = (int) points[p].x;
			y2 = (int) points[p].y;
			if (!connect) {
				x1 = x2;
				y1 = y2;
				connect = true;
			}
			g.drawLine(x1, y1, x2, y2);
			if (drawPoints) {
				g.drawLine(x2 - DATAPOINT_RAD, y2 - DATAPOINT_RAD, x2 + DATAPOINT_RAD, y2 + DATAPOINT_RAD);
				g.drawLine(x2 - DATAPOINT_RAD, y2 + DATAPOINT_RAD, x2 + DATAPOINT_RAD, y2 - DATAPOINT_RAD);
				g.drawLine(x2, y2 + DATAPOINT_RAD, x2, y2 - DATAPOINT_RAD);
				g.drawLine(x2 - DATAPOINT_RAD, y2, x2 + DATAPOINT_RAD, y2);
			}
			if (series.hlPoints[p]) {
				g.drawOval(x2 - DATAPOINT_RAD - 1, y2 - DATAPOINT_RAD - 1, 2 + 2 * DATAPOINT_RAD, 2 + 2 * DATAPOINT_RAD);
			}
			x1 = x2;
			y1 = y2;
		}
	}
} // private void paintData


/**
 * Computes the screen coordinates for the visible data series.
 */
private synchronized void computeScreenData() {
	
	if (dataComputationRunning)
		return;
	
	if (screenTooSmall())
		return;
	
	dataComputationRunning = true;
	
	computeXCoordinates();
	
	showAtLeastOneSeries = false;
	int seriesCount = dataCache.countDataSeries();
	for (int s = 0; s < seriesCount; s++) {
		
		if (!seriesSetts.getShow(s)) {
			screenDataBuffer[s].doShow = false;
			continue;
		}
		
		screenDataBuffer[s].doShow = true;
		showAtLeastOneSeries = true;
		computeScreenDataForSeries(s);		
	}
	
	dataComputationRunning = false;
}

/**
 * Compute the x coordinates in data space according to the current settings.
 */
private void computeXCoordinates() {
	
	// If the option is to use a data series, but the index is invalid, we default to dataset numbers:
	
	XAxisType xAxisType = graphSetts.getXAxisType();
	int xSerInd = -1;
	if (XAxisType.XAxis_DSNum != xAxisType) {
		xSerInd = graphSetts.getXAxisSeriesIndex();		
		if (0 > xSerInd || dataCache.countDataSeries() <= xSerInd)
			xAxisType = XAxisType.XAxis_DSNum; 		
	}
	
	// Now we can follow the secure option:
	
	int dataLen = dataCache.countDataSets();
	
	switch(xAxisType) {
		
		case XAxis_DSNum:			
			for (int i = 0; i < dataLen; i++)
				xCoordinates[i] = dataCache.getDataSet(i).getDataFileIndex();
			break;			
			
		case XAxis_DataValSimple:
			for (int i = 0; i < dataLen; i++)
				xCoordinates[i] = dataCache.getDataSet(i).getValue(xSerInd);
			break;
			
		case XAxis_DataValScaledSet:
			double factor = graphSetts.getXAxisScaleValue();
			for (int i = 0; i < dataLen; i++)
				xCoordinates[i] = dataCache.getDataSet(i).getValue(xSerInd) * factor;
			break;
			
		case XAxis_DataValTrans0to1:
			DataSeries xSer = dataCache.getDataSeries(xSerInd);
			double transfShift = xSer.getMinValue();
			double transfScale = xSer.getMaxValue() - transfShift;
			transfScale = (0 == transfScale ? 0. : 1. / transfScale);
			for (int i = 0; i < dataLen; i++)
				xCoordinates[i] = (dataCache.getDataSet(i).getValue(xSerInd) - transfShift) * transfScale;
			break;
			
		default:
			throw new Error("Unexpected x axis type");
	}
}

/**
 * Compute the screen coordinates for the specified series.
 * 
 * @param seriesIndex The cache index of the series to be computed.
 */
private void computeScreenDataForSeries(int seriesIndex) {
	
	boolean dataComputationWasRunning = dataComputationRunning;
	dataComputationRunning = true;
	
	// Preset data:
	int dataPointCount = dataCache.countDataSets();
	
	SeriesScreenData scrData = screenDataBuffer[seriesIndex];
	scrData.plotPoints = 0;	
	
	// Look at each data point of the series:
	int sp = 0;
	double x, y; DataSet ds;	
	for (int dp = 0; dp < dataPointCount; dp++) {
		
		// Get raw Y and X:
		ds = dataCache.getDataSet(dp);
		y = ds.getValue(seriesIndex);		
		y = scrData.transformer.transf(y);
		x = xCoordinates[dp];
		
		// Transform the point to screen coordinates:
		scrData.dsIndices[sp] = ds.getDataFileIndex();
		scrData.points[sp].y = y;
		scrData.points[sp].x = x;
		datScrTransform.transform(scrData.points[sp], scrData.points[sp]);
		
		// Save point index for latter sorting:
		scrData.sortedPoints[sp].val = sp;
		
		sp++;
	}
	
	// Save the number of points actually computed:
	scrData.plotPoints = sp;
	
	// Sort points for fast access when highlighting with the mouse:
	if (highlightPoints) {
		Arrays.fill(scrData.hlPoints, 0, sp, false);
		pointsByIndexComparator.setSeries(scrData);
		Arrays.sort(scrData.sortedPoints, 0, sp, pointsByIndexComparator);
	}
	
	dataComputationRunning = dataComputationWasRunning;
}


/**
 * Highlights the points around the specified point.
 * This is normally called when the mouse is moved over the plotter canvas.
 * 
 * @param sp A marker screen point.
 * @return A list of series indices on which at least one point was highlighted.
 */
public List<Integer> highlightAround(Point sp) {
	
	// If highlighting should not be done for a reason, we do not highlight anything:
	if (pointHighlightComputationRunning || dataComputationRunning || !highlightPoints) {
		List<Integer> hlSeries = Collections.emptyList();
		return hlSeries;
	}
	
	pointHighlightComputationRunning = true;
	
	List<Integer> hlSeries = new ArrayList<Integer>();
	
	// Get the rectabgle within which points will be highlighted:
	Rectangle sRect = new Rectangle(sp.x - DATAPOINT_RAD - 1, sp.y - DATAPOINT_RAD - 1,
									1 + 2 * (DATAPOINT_RAD + 1),1 + 2 * (DATAPOINT_RAD + 1));
	
	// Look for points to highlight on each series:
	SeriesScreenData series;
	for (int s = 0; s < screenDataBuffer.length; s++) {
		
		series = screenDataBuffer[s];
		
		// Skip series which are not plotted:
		if (null == series || !series.doShow)
			continue;
		
		// Clear highlight flags fopr all points of the series:
		boolean hlThisSeries = false;
		Arrays.fill(series.hlPoints, 0, series.plotPoints, false);
		
		// Find index at which the marker point would be inserted into the x-sorted series points array: 
		pointsByIndexComparator.setSeries(series);
		series.points[DataCache.CACHE_SIZE].setLocation(sp.x, sp.y);
		int mi = Arrays.binarySearch(series.sortedPoints, 0, series.plotPoints,
									 new MutableInt(DataCache.CACHE_SIZE), pointsByIndexComparator);
		if (0 > mi)  mi = -mi;
		
		// Extend array index to the left to include all points within the selection rectangle:
		int li = mi;
		if (li >= series.plotPoints) {
			li = series.plotPoints - 1;
		} else {
			while (0 <= li && series.points[series.sortedPoints[li].val].x >= sRect.x)
				li--;
			if (-1 == li) li = 0;
		}
		
		// Extend array index to the right to include all points within the selection rectangle:
		int ri = mi;
		int sRectRB = sRect.x + sRect.width;
		while (0 <= ri && ri < series.plotPoints && series.points[series.sortedPoints[ri].val].x <= sRectRB)
			ri++;
		if (ri >= series.plotPoints) ri = series.plotPoints - 1;
		
		// Now loop through the points within the determined index boundaries:
		for (int i = li; i <= ri; i++) {
		
			// Highlight a point iff it actually lies within the selection rectangle:
			if (sRect.contains(series.points[series.sortedPoints[i].val])) {
				series.hlPoints[series.sortedPoints[i].val] = true;
				hlThisSeries = true;
			}			
		}

		// If at least one point on the series was highlighted,
		// than we add the series index to the highlighted series list:
		if (hlThisSeries)
			hlSeries.add(s);
		
	}
	
	// Done:
	pointHighlightComputationRunning = false;
	return hlSeries;
}

/**
 * Computes the actual grid mesh sizes taking in account the current plot size.
 */
private void computeGridSteps() {
	
	// For horizontal grid:
	
	if (HGridType.HGrid_Simple == graphSetts.getHGridType()) {
		
		boolean hStepChanged = false;
		
		if (hGridStep != userHGridStep) {
			hGridStep = userHGridStep;
			hStepChanged = true;
		}
		
		if (hGridStep < 0.0) {
			hGridStep = -hGridStep;
			hStepChanged = true;
		}
				 
		double minHGridStep = dataViewport.height * MIN_GRIDLINE_DIST / screenSize.height;
		if (hGridStep < minHGridStep
				&& !Double.isInfinite(dataViewport.height) && 0. != dataViewport.height) {
			
			hGridStep = minHGridStep;
			double rounded = Double.parseDouble(String.format("%.3f", hGridStep));
			if (rounded < hGridStep)
				hGridStep = Double.parseDouble(String.format("%.3f", rounded + 0.001));
			else
				hGridStep = rounded;
			
			hStepChanged = true;
		}
		
		if (hStepChanged) {
			selfSettingHGridSize = true;
			graphSetts.setHGridSize(hGridStep);
			selfSettingHGridSize = false;
		}
	}
	
	// For vertical grid if it is x-axis unit-alligned:
	
	if (VGridType.VGrid_XAUnitAligned == graphSetts.getVGridType()) {
		
		boolean vStepChanged = false;
		
		if (vGridStep != userVGridStep) {
			vGridStep = userVGridStep;
			vStepChanged = true;
		}
		
		if (vGridStep < 0.0) {
			vGridStep = -vGridStep;
			vStepChanged = true;
		}
				 
		double minVGridStep = dataViewport.width * MIN_GRIDLINE_DIST / screenSize.width;
		if (vGridStep < minVGridStep
				&& !Double.isInfinite(dataViewport.width) && 0. != dataViewport.width) {
			
			vGridStep = minVGridStep;
			double rounded = Double.parseDouble(String.format("%.3f", vGridStep));
			if (rounded < vGridStep)
				vGridStep = Double.parseDouble(String.format("%.3f", rounded + 0.001));
			else
				vGridStep = rounded;
			
			vStepChanged = true;
		}
		
		if (vStepChanged) {
			selfSettingVGridSize = true;
			graphSetts.setVGridSize(vGridStep);
			selfSettingVGridSize = false;
		}
	}
	
	// For vertical grid if it is dataset-alligned:
	
	if (VGridType.VGrid_DSNumAligned == graphSetts.getVGridType()) {
		
		boolean vStepChanged = false;
		
		if (vGridStep != userVGridStep) {
			vGridStep = userVGridStep;
			vStepChanged = true;
		}
		
		double rounded = Math.rint(vGridStep);
		if (rounded != vGridStep) {
			vGridStep = rounded;
			vStepChanged = true;
		}
		
		if (vGridStep < 0.0) {
			vGridStep = -vGridStep;
			vStepChanged = true;
		}
		
		if (vGridStep == 0.0) {
			vGridStep = 1.;
			vStepChanged = true;
		}
		
		if (vStepChanged) {
			selfSettingVGridSize = true;
			graphSetts.setVGridSize(vGridStep);
			selfSettingVGridSize = false;
		}
	}
} // private void computeGridSteps()

/**
 * Map the specified point in screen coordinates into the data space.
 * 
 * @param sp A point in screen coordinates.
 * @return The corresponding data point.
 */
public Point2D.Double screenToDataPoint(Point sp) {
	
	Point2D.Double dp = new Point2D.Double();
	try {
		datScrTransform.inverseTransform(sp, dp);
	} catch(NoninvertibleTransformException e) {
		dp.setLocation(0, 0);
	}
	return dp;
}

/**
 * Updates the data to screen transform map according to the currently visible data area and screen size.
 */
private void updateDatScrTransform() {
	
	datScrTransform.setToIdentity();
	
	datScrTransform.translate(HMARGIN, (screenSize.height + VMARGIN));
	datScrTransform.scale(1, -1);
	datScrTransform.scale(screenSize.width / dataViewport.width, screenSize.height / dataViewport.height);
	datScrTransform.translate(-dataViewport.x, dataViewport.height - dataViewport.y);
}


/**
 * Reallocates the screen data buffer.
 */
private void resetScreenDataBuffer() {
	showAtLeastOneSeries = false;
	screenDataBuffer = new SeriesScreenData[dataCache.countDataSeries()];
	for (int s = 0; s < screenDataBuffer.length; s++)
		screenDataBuffer[s] = new SeriesScreenData(s); 
}


/**
 * First, recomputes the currently visible data area according to the current graph and series settings;
 * then, computes the screen coordinates for the visible data series..
 *
 */
private void updateScreenData() {
	resetDataViewport();
	computeScreenData();
}

private void updateSeriesTransformer(int seriesIndex) {
	
	if (null == screenDataBuffer)
		return;
	
	SeriesScreenData serData = screenDataBuffer[seriesIndex];
	if (null == serData)
		return;
	
	final int serInd = seriesIndex;
	
	TransformMode transformMode = seriesSetts.getTransformMode(seriesIndex);
	switch (transformMode) {
		case Transform_None:	serData.transformer = IDTransform;
								break;
		case Transform_SetVal:	final double f = seriesSetts.getScaleFactor(serInd);
								serData.transformer = new Transformer() {
									public double transf(double v) { return v * f; }
								};
								break;
		case Transform_In0to1:	DataSeries dSer = dataCache.getDataSeries(serInd);
								double serMax = dSer.getMaxValue();
								final double shift = dSer.getMinValue();
								final double scale = (0. == (serMax - shift)
														? 0. : 1. / (serMax - shift));
			serData.transformer = new Transformer() {
									public double transf(double v) { return (v - shift) * scale; }
								};
								break;
		default:				throw new Error("Unexpected series scale mode");
	}
}

/**
 * Recomputes the currently visible data area according to the current graph and series settings.
 */
private void resetDataViewport() {
	
	// If current screen is too small we dont compute:
	if (screenTooSmall())
		return;
	
	// If computation is running, we do not compute any more:
	if (dataComputationRunning)
		return;
	
	// Determine minY according to the options:
	double minY = graphSetts.getMinY();
	if (Double.isNaN(minY)) {
		
		minY = Double.MAX_VALUE;
		for (int s = 0; s < dataCache.countDataSeries(); s++) {
			
			if (!seriesSetts.getShow(s))
				continue;
			
			double v = screenDataBuffer[s].transformer.transf(dataCache.getDataSeries(s).getMinValue());
			if (v < minY)
				minY = v;
		}		
	}
	
	// Determine maxY according to the options:
	double maxY = graphSetts.getMaxY();
	if (Double.isNaN(maxY)) {		
		
		maxY = -Double.MAX_VALUE;
		for (int s = 0; s < dataCache.countDataSeries(); s++) {
			
			if (!seriesSetts.getShow(s))
				continue;
			
			double v = screenDataBuffer[s].transformer.transf(dataCache.getDataSeries(s).getMaxValue());
			if (v > maxY)
				maxY = v;
		}		
	}
	
	// Determine minX and maxX accodring to the options:
	double minX = graphSetts.getMinX();
	double maxX = graphSetts.getMaxX();
	
	// If minX or maxX are automatic and according to some data value (i.e. not dataset number):
	if ((Double.isNaN(minX) || Double.isNaN(maxX))
			&& graphSetts.getXAxisType() != XAxisType.XAxis_DSNum) {
				
		// Check that x-axis data series index is valid:
		int xAxisSerIndex = graphSetts.getXAxisSeriesIndex();
		if (0 <= xAxisSerIndex && dataCache.countDataSeries() > xAxisSerIndex) {
		
			DataSeries xSer = dataCache.getDataSeries(xAxisSerIndex);
			
			// X axis is an unscaled data series: 
			if (graphSetts.getXAxisType() == XAxisType.XAxis_DataValSimple) {
				if (Double.isNaN(minX))
					minX = xSer.getMinValue();
				if (Double.isNaN(maxX))
					maxX = xSer.getMaxValue();
				
			// X axis is a data series transformed into [0..1]:
			} else if (graphSetts.getXAxisType() == XAxisType.XAxis_DataValTrans0to1) {			
				if (Double.isNaN(minX))
					minX = 0;
				if (Double.isNaN(maxX))
					maxX = 1;
			
			// X axis is a data series scaled by a set value:
			} else if (graphSetts.getXAxisType() == XAxisType.XAxis_DataValScaledSet) {
				double scaleF = graphSetts.getXAxisScaleValue();
				if (Double.isNaN(minX))
					minX = xSer.getMinValue() * scaleF;
				if (Double.isNaN(maxX))
					maxX = xSer.getMaxValue() * scaleF;
			}			
		} // if x-axis data series index is valid
	} // if minX or maxX are automatic and according to some data value (i.e. not dataset number)
	
	// Now minX and maxX can only be NaN in one of the following cases:
	//   - x axis type is XAxis_DSNum (dataset number)
	//   - x axis series index is invalid
	//   - xSer.getMinValue or xSer.getMaxValue returned NaN
	// In all cases we do the same thing: default to dataset index:
	
	if (Double.isNaN(minX))
		minX = dataCache.getMinDataFileIndex();
	if (Double.isNaN(maxX))
		maxX = dataCache.getMaxDataFileIndex();
	
	// If the X-boundaries are equal - shift them apart:
	if (minX == maxX) {
		if (0.0 == minX)	{ minX = -0.1; maxX = 0.1; }
		if (0.0 < minX)		{ minX = 0.0;  maxX *= 1.1; }
		if (0.0 > minX)		{ minX *= 1.1; maxX = 0.0; }
	}
	
	// If the X-boundaries are the wrong way around - swap them:
	if (minX > maxX) {
		double t = maxX; maxX = minX; minX = t;
	}
	
	// If the Y-boundaries are equal - shift them apart:
	if (minY == maxY) {
		if (0.0 == minY)	{ minY = -0.1; maxY = 0.1; }
		if (0.0 < minY)		{ minY = 0.0;  maxY *= 1.1; }
		if (0.0 > minY)		{ minY *= 1.1; maxY = 0.0; }
	}
	
	// If the Y-boundaries are the wrong way around - swap them:
	if (minY > maxY) {
		double t = maxY; maxY = minY; minY = t;
	}
	
	dataViewport.setRect(minX, maxY, maxX - minX, maxY - minY);
	updateDatScrTransform();
	computeGridSteps();
} // private void resetDataViewport()


/**
 * Set the current view screen size. 
 * @param width Canvas width in pixels.
 * @param height Canvas height in pixels
 */
public void setScreenSize(int width, int height) {
	
	if (dataComputationRunning)
		return;
	
	screenSize.width = width - (HMARGIN << 1);
	screenSize.height = height - (VMARGIN << 1);
	updateDatScrTransform();
	computeScreenData();
	computeGridSteps();
}

/**
 * Gets canvas screen size (X).
 * 
 * @return Canvas screen size (X).
 */
public int getScreenWidth() {
	return screenSize.width + (HMARGIN << 1);
}

/**
 * Gets canvas screen size (Y).
 * 
 * @return Canvas screen size (Y).
 */
public int getScreenHeight() {
	return screenSize.height + (VMARGIN << 1);
}

/**
 * If cached label info is changed, the screen buffer is recreated;
 * if cached data is updated the view port and the screen data are recomputed.
 */
public void cacheEventFired(DataCache cache, CacheEvent event) {
	
	//System.out.println(event);
	
	if (cache != dataCache)
		return;
	
	if (CacheEvent.UpdateLabels == event) {
		resetScreenDataBuffer();
		for (int s = 0; s < screenDataBuffer.length; s++) {
			screenDataBuffer[s].colour = seriesSetts.getColour(s);
			updateSeriesTransformer(s);
		}
	}
	
	if (CacheEvent.UpdateData == event) {
		updateScreenData();
	}
}

/**
 * Dispatches settings change events.
 */
public void settingHasChanged(ObservableSettings settings, Object info) {
	
	//System.out.println(settings.getClass().getName() + ": " + info);
	
	if (null == info || null == settings)
		return;
	
	if ((settings instanceof DataSeriesSettings) && (info instanceof String)) {
		settingHasChanged((DataSeriesSettings) settings, (String) info);
		return;
	}
	
	if ((settings instanceof GraphSettings) && (info instanceof String)) {
		settingHasChanged((GraphSettings) settings, (String) info);
		return;
	}
}

/**
 * Calls the neccesary recoputations when seties settings are changed.
 * @param settings Series settings.
 * @param info Change event info.
 */
public void settingHasChanged(DataSeriesSettings settings, String info) {
	
	if (null == info || null == settings)
		return;
	
	if (info.equals("load")) {		
		for (int s = 0; s < screenDataBuffer.length; s++) {
			screenDataBuffer[s].colour = settings.getColour(s);
			updateSeriesTransformer(s);
		}
		updateScreenData();
		return;
	}	
	
	if (info.startsWith("Show")) {
		updateScreenData();	
	}
	
	if (info.startsWith("TransformMode.")) {
		int affectedSeries = Integer.parseInt(info.substring(info.lastIndexOf(".") + 1));
		updateSeriesTransformer(affectedSeries);
		updateScreenData();	
	}
	
	if (info.startsWith("Colour.")) {		
		int affectedSeries = Integer.parseInt(info.substring(info.lastIndexOf(".") + 1));
		screenDataBuffer[affectedSeries].colour = settings.getColour(affectedSeries);		
	}
	
	if (info.startsWith("ScaleFactor.")) {
		int affectedSeries = Integer.parseInt(info.substring(info.lastIndexOf(".") + 1));
		if (TransformMode.Transform_SetVal == seriesSetts.getTransformMode(affectedSeries)) {
			updateSeriesTransformer(affectedSeries);
			updateScreenData();
		}
	}
}

/**
 * Calls the neccesary recoputations when graph settings are changed.
 * @param settings Graph settings.
 * @param info Change event info.
 */
public void settingHasChanged(GraphSettings settings, String info) {
	
	if (null == info || null == settings || settings != this.graphSetts)
		return;
	
	if (info.equals("MinY") || info.equals("MinX")
			|| info.equals("MaxY") || info.equals("MaxX")
			|| info.equals("load")) {
		updateScreenData();
	}
	
	if (info.equals("VGridType") || info.equals("VGridSize") || info.equals("load")) {
		if (!selfSettingVGridSize) {
			userVGridStep = graphSetts.getVGridSize();
			computeGridSteps();
		}
	}
	
	if (info.equals("HGridType") || info.equals("HGridSize") || info.equals("load")) {
		if (!selfSettingHGridSize) {
			userHGridStep = graphSetts.getHGridSize();
			computeGridSteps();
		}
	}
	
	if (info.equals("XAxisType") || info.equals("XAxisSeriesIndex") || info.equals("XAxisScaleValue")
			|| info.equals("load")) {
		updateScreenData();
	}
	
	if (info.equals("HighlightDataPoints") || info.equals("load")) {
		highlightPoints = settings.getHighlightDataPoints();
		highlightAround(new Point(-1, -1));
	}
}

/**
 * For holding mutable ints as objects.
 */
private class MutableInt {
	/*package*/ int val = -1;
	/*package*/ MutableInt() { this.val = -1; }
	/*package*/ MutableInt(int v) { this.val = v; }
}

/**
 * A data structure to hold the locally cached plot data for a data series.
 */
private class SeriesScreenData {

	/*package*/ Color colour = Color.BLACK;
	/*package*/ int series = -1;
	/*package*/ Point2D.Double[] points = new Point2D.Double[DataCache.CACHE_SIZE + 1];
	/*package*/ MutableInt[] sortedPoints = new MutableInt[DataCache.CACHE_SIZE];
	/*package*/ boolean[] hlPoints = new boolean[DataCache.CACHE_SIZE];
	/*package*/ int[] dsIndices = new int[DataCache.CACHE_SIZE];
	/*package*/ int plotPoints = 0;
	/*package*/ boolean doShow = false; 
	/*package*/ Transformer transformer = IDTransform;
	
	/*package*/ SeriesScreenData(int series) {
		this.series = series;
		for(int i = 0; i < DataCache.CACHE_SIZE; i++) {
			points[i] = new Point2D.Double();
			sortedPoints[i] = new MutableInt();
		}
		points[DataCache.CACHE_SIZE] = new Point2D.Double();
	}

} // private class SeriesScreenData


/**
 * Used in order to compare points referenced by their index in {@link SeriesScreenData#points};
 * the comparison is by x-xoordinates. 
 */
private class PointsByIndexComparator implements Comparator<MutableInt> {
	private SeriesScreenData series = null;
	/*package*/ void setSeries(SeriesScreenData series) { this.series = series; }
	public int compare(MutableInt pi1, MutableInt pi2) {
		Point2D.Double p1 = series.points[pi1.val];
		Point2D.Double p2 = series.points[pi2.val];
		if (p1.x < p2.x)
			return -1;
		if (p1.x > p2.x)
			return 1;
		return 0;
    }	
} // private class PointsByIndexComparator

/**
 * Used to encapsulate data series points translation routines.
 */
private interface Transformer {
	public double transf(double v);
}
private static Transformer IDTransform = new Transformer(){public double transf(double v){return v;}};

} // public class Plotter
